"""
test_elements contains the test cases for the SQcircuit elements functionalities.
"""
import pytest
from SQcircuit.elements import *
import SQcircuit as sq

#######################################
# Capacitor Tests
#######################################


def test_capacitorErrors():
    error = "The input unit for the capacitor is not correct. Look at the documentation for the correct input " \
            "format."
    with pytest.raises(ValueError, match=error):
        Capacitor(10, "H")


def test_capacitorEnergy():
    cap = Capacitor(10, "GHz")
    assert cap.energy() == 10
    # change the default frequency of the SQcirucit
    sq.unit.setFreq("MHz")
    # check back the energy
    assert cap.energy() == 10 * 1000
    # set back the default frequency back to GHz
    sq.unit.setFreq("GHz")
    # check the energy functionality from setting the value
    val = cap.value()
    cap2 = Capacitor(val, "F")
    assert cap2.energy() == 10


def test_capacitorQ():
    cap = Capacitor(10, "GHz", Q=1)
    assert cap.Q(-1) == 1
    assert cap.Q(10) == 1
    assert cap.Q(0) == 1

    cap = Capacitor(10, "GHz", Q=1e7)
    assert cap.Q(-1) == 1e7
    assert cap.Q(10) == 1e7
    assert cap.Q(0) == 1e7

    Q = lambda omega: omega ** 2
    cap = Capacitor(10, "GHz", Q=Q)
    assert cap.Q(2) == 4

#######################################
# Inductor Tests
#######################################


def test_inductorErrors():
    error = "The input unit for the inductor is not correct. Look at the documentation for the correct input " \
            "format."
    with pytest.raises(ValueError, match=error):
        Inductor(10, "F")


def test_inductorEnergy():
    ind = Inductor(10, "GHz")
    assert ind.energy() == 10
    # change the default frequency of the SQcirucit
    sq.unit.setFreq("MHz")
    # check back the energy
    assert ind.energy() == 10 * 1000
    # set back the default frequency back to GHz
    sq.unit.setFreq("GHz")
    # check the energy functionality from setting the value
    val = ind.value()
    ind2 = Inductor(val, "H")
    assert ind2.energy() == 10


def test_inductorQ():
    ind = Inductor(10, "GHz", Q=1)
    assert ind.Q(-1, 0) == 1
    assert ind.Q(10, 12) == 1
    assert ind.Q(0, 11) == 1

    ind = Inductor(10, "GHz", Q=1e7)
    assert ind.Q(-1, 5) == 1e7
    assert ind.Q(10, 6) == 1e7
    assert ind.Q(0, 1) == 1e7

    ind = Inductor(10, "GHz")
    assert np.isclose(ind.Q(2 * np.pi * 0.5e9, 10), 500e6)
    assert np.isclose(ind.Q(2 * np.pi * 0.5e9, 1), 500e6)

    Q = lambda omega, T: omega ** 2
    ind = Inductor(10, "GHz", Q=Q)
    assert ind.Q(2, 10) == 4


#######################################
# Josephson Junction Tests
#######################################

def test_junctionErrors():
    error = "The input unit for the Josephson Junction is not correct. Look at the documentation for the" \
                    "correct input format."
    with pytest.raises(ValueError, match=error):
        Junction(10, "F")


def test_junctionY():
    yFunc = lambda omega, T: omega*T
    JJ = Junction(10, "GHz", Y=yFunc)
    assert JJ.Y(10, 2) == 20




class Units:
    """
    class that contains the units and physical constant for the SQcircuit
    """

    def __init__(self):
        # Henry units
        self.henryList = {'H': 1.0, 'mH': 1.0e-3, 'uH': 1.0e-6, 'nH': 1.0e-9, 'pH': 1.0e-12, 'fH': 1.0e-15}

        # Farad units
        self.faradList = {'F': 1, 'mF': 1.0e-3, 'uF': 1.0e-6, 'nF': 1.0e-9, 'pF': 1.0e-12, 'fF': 1.0e-15}

        # frequency unit list
        self.freqList = {'Hz': 1.0, 'kHz': 1.0e3, 'MHz': 1.0e6, 'GHz': 1.0e9, 'THz': 1.0e12}

        # reduced Planck constant
        self.hbar = 1.0545718e-34

        # magnetic flux quantum
        self.Phi0 = 2.067833e-15

        # electron charge
        self.e = 1.6021766e-19

        self.k_B = 1.38e-23

        # main frequency unit of the SQcircuit
        self.freq = self.freqList["GHz"]

    def setFreq(self, fUnit):
        """
        changes the main frequency unit of the SQcircuit
        inputs:
            -- fUnit: the desired frequency unit
        """
        assert fUnit in self.freqList, "The input format is not correct."

        self.freq = self.freqList[fUnit]


# The global units of the SQcircuit.
unit = Units()



"""
test_qubits contains the test cases for well-known qubits.
"""

from SQcircuit.tests.conftest import QubitTest


class TestZeroPi(QubitTest):
    """
    class for testing the zero-pi qubit
    """
    @classmethod
    def setup_class(cls):
        cls.fileName = "zeroPi_1"


class TestInductivelyShunted(QubitTest):
    """
    class for testing the zero-pi qubit
    """
    @classmethod
    def setup_class(cls):
        cls.fileName = "inductivelyShunted_1"


class TestFluxonium(QubitTest):
    """
    class for testing the zero-pi qubit
    """
    @classmethod
    def setup_class(cls):
        cls.fileName = "Fluxonium_1"


class TestFluxonium2(QubitTest):
    """
    class for testing the zero-pi qubit
    """
    @classmethod
    def setup_class(cls):
        cls.fileName = "Fluxonium_2"


class TestTransmon(QubitTest):
    """
    class for testing the zero-pi qubit
    """
    @classmethod
    def setup_class(cls):
        cls.fileName = "Transmon_1"




import sys
import os

import numpy as np
import matplotlib.pyplot as plt

sys.path.insert(0, os.path.abspath('../..'))

import SQcircuit as sq

tests = {"zeroPi": False,
         "inductivelyShunted": False,
         "Fluxonium": False,
         "Transmon": False,
         "tunableTransmon": False,
         "isolatedIsland": False
         }

#######################################
# zero-pi qubit
#######################################

if tests["zeroPi"]:
    loop1 = sq.Loop()

    C = sq.Capacitor(0.15, "GHz")
    CJ = sq.Capacitor(10, "GHz")
    JJ = sq.Junction(5, "GHz", loops=[loop1])
    L = sq.Inductor(0.13, "GHz", loops=[loop1])

    elements = {(0, 1): [CJ, JJ],
                (0, 2): [L],
                (3, 0): [C],
                (1, 2): [C],
                (1, 3): [L],
                (2, 3): [CJ, JJ],
                }

    # cr is an object of Qcircuit
    cr = sq.Circuit(elements)

    cr.truncationNumbers([25, 1, 25])

    sweep1 = sq.Sweep(cr, numEig=5)

    phi = np.linspace(0, 1, 50) * 2 * np.pi

    sweep1.sweepFlux([loop1], [phi], plotF=True, toFile='data/zeroPi_1')
    # sweep1.sweepFlux([loop1], [phi], plotF=True)

#######################################
# inductively shunted qubit
#######################################

if tests["inductivelyShunted"]:
    loop1 = sq.Loop()

    # define the circuit’s elements
    C_r = sq.Capacitor(20.3, "fF")
    C_q = sq.Capacitor(5.3, "fF")
    L_r = sq.Inductor(15.6, "nH")
    L_q = sq.Inductor(386, "nH", loops=[loop1])
    L_s = sq.Inductor(4.5, "nH", loops=[loop1])
    JJ = sq.Junction(6.2, "GHz", loops=[loop1])

    # define the circuit
    elements = {(0, 1): [C_r],
                (1, 2): [L_r],
                (0, 2): [L_s],
                (2, 3): [L_q],
                (0, 3): [JJ, C_q]}

    cr = sq.Circuit(elements)

    cr.truncationNumbers([1, 9, 23])

    sweep1 = sq.Sweep(cr, numEig=10)

    phi = np.linspace(-0.1, 0.6, 50) * 2 * np.pi

    sweep1.sweepFlux([loop1], [phi], plotF=True, toFile='data/inductivelyShunted_1')
    # sweep1.sweepFlux([loop1], [phi], plotF=True)

#######################################
# Fluxonium
#######################################

if tests["Fluxonium"]:
    # fluxonium with charge island( added by Yudan)

    loop1 = sq.Loop()
    C = sq.Capacitor(11, 'fF', Q=1e6)
    Cg = sq.Capacitor(0.5, 'fF', Q=1e6)
    L = sq.Inductor(1, 'GHz', Q=500e6, loops=[loop1])
    JJ = sq.Junction(3, 'GHz', cap=C, A=5e-7, loops=[loop1])

    circuitElements = {
        (0, 1): [Cg],
        (1, 2): [L, JJ],
        (0, 2): [Cg]
    }

    cr = sq.Circuit(circuitElements)

    cr.truncationNumbers([30, 1])

    sweep1 = sq.Sweep(cr, numEig=10)

    phi = np.linspace(0.0, 1.0, 50) * 2 * np.pi

    sweep1.sweepFlux([loop1], [phi], plotF=True, toFile='data/Fluxonium_1')
    # sweep1.sweepFlux([loop1], [phi], plotF=True)

    # standard fluxonium with loss calculation

    loop1 = sq.Loop(A=1e-6)

    C = sq.Capacitor(3.6, 'GHz', Q=1e6)
    L = sq.Inductor(0.46, 'GHz', Q=500e6, loops=[loop1])
    JJ = sq.Junction(10.2, 'GHz', cap=C, A=1e-7, x=3e-06, loops=[loop1])

    circuitElements = {
        (0, 1): [L, JJ],
    }

    cr = sq.Circuit(circuitElements)

    cr.truncationNumbers([200])

    sweep1 = sq.Sweep(cr, numEig=10, properties=["efreq", "loss"])

    phi = np.linspace(0.1, 0.9, 50) * 2 * np.pi

    sweep1.sweepFlux([loop1], [phi], plotF=True, toFile='data/Fluxonium_2')


if tests["Transmon"]:

    C = sq.Capacitor(20, 'GHz', Q=1e6)
    JJ = sq.Junction(15.0, 'GHz', A=1e-7)

    circuitElements = {
        (0, 1): [C, JJ]
    }

    cr = sq.Circuit(circuitElements)

    cr.truncationNumbers([30])

    ng = np.linspace(-0.05, 0.05, 200)

    # cr.chargeNoise(mode=1, A=1e-5)

    sweep1 = sq.Sweep(cr, numEig=6, properties=["efreq", "loss"])

    sweep1.sweepCharge([1], [ng], plotF=True, toFile='data/Transmon_1')

    plt.semilogy(ng, 1/sweep1.dec['charge'])
    plt.show()


#######################################
# isolatedIsland
#######################################

if tests["isolatedIsland"]:

    # fluxonium with charge island( added by Yudan)

    loop1 = sq.Loop()
    C = sq.Capacitor(11, 'fF', Q=1e6)
    Cg = sq.Capacitor(0.5, 'fF', Q=1e6)
    L = sq.Inductor(1, 'GHz', Q=500e6, loops=[loop1])
    JJ = sq.Junction(3, 'GHz', cap=C, A=5e-7, loops=[loop1])

    circuitElements = {
        (1, 2): [Cg],
        (0, 1): [JJ, L],
        (2, 3): [Cg],
        (0, 2): [Cg],
    }

    cr = sq.Circuit(circuitElements)

    print(cr.W)

    cr.truncationNumbers([30, 1, 1])

    sweep1 = sq.Sweep(cr, numEig=10)

    phi = np.linspace(0.0, 1.0, 50) * 2 * np.pi

    # sweep1.sweepFlux([loop1], [phi], plotF=True, toFile='data/Fluxonium_1')
    # sweep1.sweepFlux([loop1], [phi], plotF=True)


"""
storage.py contains the classes for storing the data
"""
import dill as pickle
import matplotlib.pyplot as plt
import numpy as np


class SQdata:
    """
    Class for processing and representing the data related to a circuit.
    """
    # type of the data
    type = None
    # circuit of the data
    cr = None
    # parameters related to data
    params = None
    # grid related to data
    grid = None
    # eigenfrequencies of the circuit
    efreq = None
    # the decayRates
    dec = None

    def save(self, toFile: str):
        with open(toFile, 'wb') as outp:
            pickle.dump(self, outp, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(fromFile: str):
        with open(fromFile, 'rb') as inp:
            return pickle.load(inp)

    def plot(self, typeP):
        assert len(self.efreq.shape) <= 2, "SQcircuit can only plot 1D grid data for now"

        plt.figure(figsize=(6.5, 5), linewidth=1)
        plt.tight_layout()
        # plt.title("Energy Spectrum", fontsize=18)
        for i in range(self.numEig):
            plt.plot(self.grid[0] / 2 / np.pi, self.efreq[i, :] - self.efreq[0, :],
                     linewidth=2.2)
        if typeP == "flux":
            plt.xlabel(r"$\Phi_{ext}/\Phi_0$", fontsize=18)
        elif typeP == "charge":
            plt.xlabel(r"$n_g$", fontsize=18)
        plt.ylabel(r"$f_i-f_0$[GHz]", fontsize=18)
        plt.show()
